using System;
using System.Collections;
using System.Drawing;

namespace ECAL
{
	/// <summary>
	/// Summary description for Environment.
	/// </summary>
	public class Environment
	{
		/// <summary>
		/// Group of robots in the ground
		/// </summary>
		public ArrayList MobileRobots = new ArrayList();
		/// <summary>
		/// Group of obstacles in the ground
		/// </summary>
		public ArrayList Obstacles = new ArrayList();
		/// <summary>
		/// Group of walls in the ground
		/// </summary>
		public ArrayList Walls = new ArrayList();
		/// <summary>
		/// Group of goals in the ground for robot(s)
		/// </summary>
		public ArrayList Goals = new ArrayList();
		/// <summary>
		/// Width of the ground
		/// </summary>
		public double GroundWidth = 30.0;
		/// <summary>
		/// Height of the ground
		/// </summary>
		public double GroundHeight = 30.0;
		/// <summary>
		/// Random number to be used for generating random values
		/// </summary>
		Random RandomNum;
		/// <summary>
		/// List of errors to be reported
		/// </summary>
		public ArrayList RepErrors = new ArrayList();
		/// <summary>
		/// List of rectangle areas to be reported
		/// </summary>
		public ArrayList RepRectangleArea = new ArrayList();
		/// <summary>
		/// Values of cells in the environment
		/// Cells are counted from 0
		/// </summary>
		public double [,] Cells = new double[50,50];
		/// <summary>
		/// Width or Height of cells in the environment
		/// It is considered that the cells are square
		/// </summary>
		public double CellsWidth = 1.0;
		/// <summary>
		/// Creatures states in the environment (dead or alive)
		/// This part is designed for bacteria simulation (cellular automata)
		/// </summary>
		public double [,] Creatures = new double[50,50];
		/// <summary>
		/// Next values for creatures of the environment (dead or alive)
		/// These values are applied when performing the perform action
		/// </summary>
		public double [,] NextCreatures = new double[50,50];
		private Bitmap Image;
		public ArrayList IPAgents = new ArrayList();

		public Environment(Random rnd, Bitmap img)
		{
			RandomNum = rnd;
			Image = img;

			ImageAgent TempIPAgent;
			
			TempIPAgent = new ImageAgent(Image);
			TempIPAgent.X = 20;
			TempIPAgent.Y = 20;
			IPAgents.Add(TempIPAgent);
			/*
			TempIPAgent = new ImageAgent(Image);
			TempIPAgent.X = 20;
			TempIPAgent.Y = 0;
			IPAgents.Add(TempIPAgent);
			
			TempIPAgent = new ImageAgent(Image);
			TempIPAgent.X = 40;
			TempIPAgent.Y = 0;
			IPAgents.Add(TempIPAgent);
			
			TempIPAgent = new ImageAgent(Image);
			TempIPAgent.X = 60;
			TempIPAgent.Y = 0;
			IPAgents.Add(TempIPAgent);
			
			TempIPAgent = new ImageAgent(Image);
			TempIPAgent.X = 80;
			TempIPAgent.Y = 0;
			IPAgents.Add(TempIPAgent);
			*/
		}

		public void MakePicBlackWhite()
		{
			Color PixColor;

			for (int i=0 ; i < Image.Size.Width ; i++)
				for (int j=0 ; j < Image.Size.Height ; j++)
				{
					PixColor = Image.GetPixel(i, j);
					Image.SetPixel(i, j, Color.FromArgb((int)(255*PixColor.GetBrightness()), (int)(255*PixColor.GetBrightness()), (int)(255*PixColor.GetBrightness())));
				}
		}

		private int CreatureNeighborsCount(int i, int j)
		{
			int Result = 0;

			if (Creatures[i-1,j] > 0.0)
				Result++;
			
			if (Creatures[i+1,j] > 0.0)
				Result++;
			
			if (Creatures[i,j-1] > 0.0)
				Result++;

			if (Creatures[i,j+1] > 0.0)
				Result++;

			return (Result);
		}
/*
		private bool CreatureSuffocates(int i, int j)
		{
			if ((Creatures[i-1,j] > 0.0)
				&&
				(Creatures[i+1,j] > 0.0)
				&&
				(Creatures[i,j-1] > 0.0)
				&&
				(Creatures[i,j+1] > 0.0))
			{
				return(true);
			}
			else
			{
				return(false);
			}
		}
*/
		/*
		private bool CreatureIsAlone(int i, int j)
		{
			if ((Creatures[i-1,j] > 0)
				||
				(Creatures[i+1,j] > 0)
				||
				(Creatures[i,j-1] > 0)
				||
				(Creatures[i,j+1] > 0))
			{
				return(false);
			}
			else
			{
				return(true);
			}
		}
*/
		private void MakeCreatureAround(int i,int j)
		{
			int Number = RandomNum.Next(1,5);

			switch (Number)
			{
				case 1:
					NextCreatures[i+1,j] = NextCreatures[i+1,j] + 1;
					break;
				case 2:
					NextCreatures[i,j+1] = NextCreatures[i,j+1] + 1;
					break;
				case 3:
					NextCreatures[i-1,j] = NextCreatures[i-1,j] + 1;
					break;
				case 4:
					NextCreatures[i,j-1] = NextCreatures[i,j-1] + 1;
					break;
				default:
					Number = 0;
					break;
			}
		}

		public void DecideForCreatures()
		{
			double RND;

			for (int i=1 ; i < 49 ; i++)
				for (int j=1 ; j < 49 ; j++)
				{
					switch (CreatureNeighborsCount(i,j))
					{
						case 0:
							NextCreatures[i,j] = 0.0;
							break;
						case 1:
							RND = RandomNum.NextDouble();
							if (RND < 0.85)
								MakeCreatureAround(i,j);
							break;
						case 2:
							RND = RandomNum.NextDouble();
							if (RND < 0.5)
								NextCreatures[i,j] = 0.0;
							break;
						case 3:
							RND = RandomNum.NextDouble();
							if (RND < 0.85)
								NextCreatures[i,j] = 0.0;
							break;
						case 4:
							NextCreatures[i,j] = 0.0;
							break;
					}
				}
		}

		public void DoForCreatures()
		{
			for (int i=1 ; i < 49 ; i++)
				for (int j=1 ; j < 49 ; j++)
				{
					Creatures[i,j] = NextCreatures[i,j];
				}
		}

		/// <summary>
		/// Routine for declaring the robots position
		/// So, it is up to environment to mark cells or do anything needed
		/// </summary>
		/// <param name="robot"></param>
		public void DeclareRobotsPosition(MobileRobot robot)
		{
			int CellX = (int)(robot.AcX / CellsWidth);
			int CellY = (int)(robot.AcY / CellsWidth);

//			if ((CellX >= 0) && (CellY >= 0))
			Cells[CellX, CellY] = Cells[CellX, CellY] + 1.0;
		}

		/// <summary>
		/// Calculates the distance between two points
		/// </summary>
		/// <param name="x1"></param>
		/// <param name="y1"></param>
		/// <param name="x2"></param>
		/// <param name="y2"></param>
		/// <returns></returns>
		public double Distance(double x1, double y1, double x2, double y2)
		{
			return(Math.Sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)));
		}

		public void RobotsInCircle(
			double centreX,
			double centreY,
			double radius,
			ref ArrayList visibleRobots)
		{
//			visibleRobots = new ArrayList();
			visibleRobots.Clear();
			foreach(MobileRobot robot in MobileRobots)
			{
				double Dist = Distance(centreX, centreY, robot.AcX, robot.AcY);
				if (Dist <= radius)
					visibleRobots.Add(robot);
			}
		}

		public void NormalRandomNumber(out double num1, out double num2)
		{
			double x1, x2, w;//, y1, y2;
 
			do 
			{
				x1 = 2.0 * RandomNum.NextDouble() - 1.0;
				x2 = 2.0 * RandomNum.NextDouble() - 1.0;
				w = x1 * x1 + x2 * x2;
			} while ( w >= 1.0 );

			w = Math.Sqrt((-2.0 * Math.Log(w)) / w);
			num1 = x1 * w;
			num2 = x2 * w;
		}

		/// <summary>
		/// Calculates total error for all robots (as distance)
		/// </summary>
		/// <returns>Value of total error</returns>
		public double TotalError()
		{
			double Result = 0.0;
			foreach(MobileRobot rob in MobileRobots)
			{
				Result = Result + Distance(rob.ImX, rob.ImY, rob.AcX, rob.AcY);
			}

			return Result;
		}

		public double TotalRobotsRectangleArea()
		{
			double Result = 0.0;
			double MinX = this.GroundWidth;
			double MaxX = 0.0;
			double MinY = this.GroundHeight;
			double MaxY = 0.0;

			foreach(MobileRobot rob in MobileRobots)
			{
				if (rob.AcX < MinX)
					MinX = rob.AcX;
				if (rob.AcX > MaxX)
					MaxX = rob.AcX;
				if (rob.AcY < MinY)
					MinY = rob.AcY;
				if (rob.AcY > MaxY)
					MaxY = rob.AcY;
			}

			Result = (MaxX - MinX) * (MaxY - MinY);
			return (Result);
		}

		/// <summary>
		/// Cleans the environment entirely
		/// </summary>
		public void CleanUpEntirely()
		{
			MobileRobots.Clear();
			Obstacles.Clear();
			Walls.Clear();
			Goals.Clear();
			RepErrors.Clear();
		}
	}
}
