using System;
using System.Collections;

namespace ECAL
{
	public enum RobotCorrectionMethods
	{
		None,
		Average,
		WightedAverage,
		OWA
	}

	public enum RobotWorkingModes
	{
		/// <summary>
		/// Follows the goal in th eenvironment
		/// </summary>
		TargetFollowing,
		CompleteCoverage,
		Leader,
		WallFollowing,
		/// <summary>
		/// Follows the environment such that finds the corners of the environment
		/// </summary>
		EnvironmentFollowing
	}

	public enum ObstacleAvoidanceMethods
	{
		/// <summary>
		/// By potential, considers all obstacles in the environment
		/// </summary>
		Potential,
		/// <summary>
		/// Uses sensory data to avoid obstacles locally
		/// </summary>
		Sensor
	}

	/// <summary>
	/// Summary description for MobileRobot.
	/// </summary>
	public class MobileRobot
	{
		/// <summary>
		/// Position of the robot on the ground as it imagines
		/// </summary>
		public double ImX;
		public double ImY;
		/// <summary>
		/// Actual position of the robot on the ground
		/// </summary>
		public double AcX;
		public double AcY;
		/// <summary>
		/// Position to be added for next step
		/// </summary>
		private double AddingX;
		private double AddingY;
		/// <summary>
		/// Radius of the robot as a circle
		/// </summary>
		public double Radius;
		/// <summary>
		/// Load of the robot
		/// </summary>
		public double Load;
		/// <summary>
		/// Environment which the robot is located in
		/// </summary>
		private Environment Environment;
		/// <summary>
		/// Random number of the program
		/// </summary>
		Random RandomNum;
		/// <summary>
		/// The radius of the circle the robot is able to see around and
		/// communicate with robots inside it
		/// </summary>
		double SeeRadius = 3;
		public double Speed = 1.0;
		public RobotCorrectionMethods CorrectionMethod = RobotCorrectionMethods.None;
		public RobotWorkingModes WorkMode = RobotWorkingModes.TargetFollowing;
		/// <summary>
		/// Length of spring between this robot and other visible robots around it
		/// </summary>
		public double SpringLength = 3.0;
		/// <summary>
		/// Radius of a circle which specifies the visible located robots by this robot
		/// </summary>
		public double SeeDomain = 5.0;
		/// <summary>
		/// The method which the robot uses for obstacle avoidance
		/// </summary>
		public ObstacleAvoidanceMethods ObsAvoidanceMethod = ObstacleAvoidanceMethods.Sensor;
		/// <summary>
		/// Maximum distance that a wall can be seen by sensors of the robot
		/// </summary>
		public double MaxWallVisDist = 0.5;
		/// <summary>
		/// Multiplier that is multiplied by the value of error for when
		/// error is generated while movements
		/// </summary>
		public double ErrorMultiplier = 0.0;//0.1;
		/// <summary>
		/// Maximum value of force applied to a robot because of being near to wall
		/// This saturation is considered for the robot in order to avoid large
		/// values of movements
		/// </summary>
		public double MaxForceFromWall = 1.0;
		/// <summary>
		/// Number of times the robot has changed its cell
		/// This variable is considered because of continous movements of the robot
		/// and we have used this variable to count number of discrete changes of
		/// cells the robot makes
		/// </summary>
		public int NumOfCellChanges = 0;
		/// <summary>
		/// Text variable just for debug and will be removed later
		/// </summary>
		public string DebugText = "";
		/// <summary>
		/// Group of visible robots that this robot can see
		/// </summary>
		public ArrayList VisibleRobots = new ArrayList();

		public MobileRobot(Environment environment, Random rnd)
		{
			Environment = environment;
			RandomNum = rnd;
		}

		/// <summary>
		/// Routine to check if the robot reached the goal or not
		/// </summary>
		/// <returns>true if robot is on goal and false if robot not on goal yet</returns>
		private bool ReachedGoal()
		{
			Goal theGoal = (Goal)Environment.Goals[0];
			double Distance;
			Distance = Math.Sqrt(Math.Pow(theGoal.X-this.AcX,2) + Math.Pow(theGoal.Y-this.AcY,2));

			if (Distance < (theGoal.Radius+this.Radius))
				return true;
			else
				return false;
		}

		private bool PointInRectangle(
			double pointX,
			double pointY,
			double rectX,
			double rectY,
			double rectWidth,
			double rectHeight)
		{
			bool PointInRect = true;
			PointInRect = pointX > rectX;
			PointInRect = PointInRect && (pointX < (rectX + rectWidth));
			PointInRect = PointInRect && (pointY > rectY);
			PointInRect = PointInRect && (pointY < (rectY + rectHeight));

			return (PointInRect);
		}

		/// <summary>
		/// Calculates min distance of a point from a rectangle
		/// </summary>
		/// <param name="pointX"></param>
		/// <param name="pointY"></param>
		/// <param name="rectX"></param>
		/// <param name="rectY"></param>
		/// <param name="rectWidth"></param>
		/// <param name="rectHeight"></param>
		/// <param name="xValue">Value of X</param>
		/// <param name="yValue">Value of Y</param>
		private void DistanceFromRectangle(
			double pointX,
			double pointY,
			double rectX,
			double rectY,
			double rectWidth,
			double rectHeight,
			out double xValue,
			out double yValue)
		{
			bool PointInRect = true;
			PointInRect = pointX > rectX;
			PointInRect = PointInRect && (pointX < (rectX + rectWidth));
			PointInRect = PointInRect && (pointY > rectY);
			PointInRect = PointInRect && (pointY < (rectY + rectHeight));
			if (PointInRect)
				PointInRect = false;

			//			double Distance = 0;
			double XResult = 0;
			double YResult = 0;

			if (pointX < rectX) // W or NW or SW
			{
				if (pointY > rectY) // NW
				{
					//					Distance = Math.Sqrt(Math.Pow(rectX-pointX,2) + Math.Pow(rectY-pointY,2));
					XResult = rectX - pointX;
					YResult = rectY - pointY;
				}
				else // W or SW
				{
					if (pointY < (rectY-rectHeight)) // SW
					{
						//						Distance = Math.Sqrt(Math.Pow(rectX-pointX,2) + Math.Pow(rectY-rectHeight-pointY,2));
						XResult = rectX - pointX;
						YResult = rectY - rectHeight - pointY;
					}
					else // W
					{
						//						Distance = pointX - rectX; // the value will be negative
						XResult = rectX - pointX;
						YResult = 0;
					}
				}
			}
			else // N or S or NE or E or SE
			{
				if (pointX > (rectX + rectWidth)) // E or NE or SE
				{
					if (pointY > rectY) // NE
					{
						//						Distance = Math.Sqrt(Math.Pow(rectX + rectWidth - pointX,2) + Math.Pow(rectY - pointY,2));
						XResult = rectX + rectWidth - pointX;
						YResult = rectY - pointY;
					}
					else // E or SE
					{
						if (pointY < (rectY - rectHeight)) // SE
						{
							//							Distance = Math.Sqrt(Math.Pow(rectX+rectWidth-pointX,2) + Math.Pow(rectY-rectHeight-pointY,2));
							XResult = rectX + rectWidth - pointX;
							YResult = rectY - rectHeight - pointY;
						}
						else // E
						{
							//							Distance = pointX - rectX - rectWidth /*- this.Radius*/;
							XResult = rectX + rectWidth - pointX;
							YResult = 0;
						}
					}
				}
				else // N or S
				{
					if (pointY > rectY) // N
					{
						XResult = 0;
						YResult = rectY - pointY;
					}
					else // S
					{
						XResult = 0;
						YResult = rectY - rectHeight - pointY;
					}
				}
			}

			xValue = XResult;
			yValue = YResult;
		}

		private void ForceFromWall(Wall wall, out double xForce, out double yForce)
		{
			double xForceResult = 0;
			double yForceResult = 0;

			double XDist = 0;
			double YDist = 0;
			DistanceFromRectangle(
				this.AcX,
				this.AcY,
				wall.X,
				wall.Y,
				wall.Width,
				wall.Height,
				out XDist,
				out YDist);

			/// Nearest distance from center of robot to the wall
			double CentreDistance = Math.Sqrt(Math.Pow(XDist,2) + Math.Pow(YDist,2));
			/// Distance from surface of robot to the wall
			double SurfaceDistance = CentreDistance - this.Radius;
			
			double Force;

			if ((SurfaceDistance >= 0.0) && (SurfaceDistance <= 5*this.Radius))
			{
				XDist = (XDist * SurfaceDistance) / CentreDistance;
				YDist = (YDist * SurfaceDistance) / CentreDistance;

				Force = -(1/SurfaceDistance) * wall.Load * this.Load;
			}
			else
			{
				Force = 0.0;
			}

			if (Math.Abs(Force) > MaxForceFromWall)
				Force = Math.Sign(Force) * MaxForceFromWall;

			xForceResult = (XDist/SurfaceDistance) * Force;
			yForceResult = (YDist/SurfaceDistance) * Force;

			xForce = xForceResult;
			yForce = yForceResult;
		}

		/// <summary>
		/// Calculates the force of a spring
		/// </summary>
		/// <param name="spLength">Length of the spring</param>
		/// <param name="spDeltaLength">Difference in length of the spring</param>
		/// <param name="spK">K multiplier of the spring</param>
		/// <returns></returns>
		private double SpringForce(double spLength, double spDeltaLength, double spK)
		{
			double DeltaDist = spDeltaLength;
			
			/// To make sure that the spring is not strenghed too much
			/// this makes a limitation on the amount of generated force
			if (Math.Abs(spDeltaLength) > 0.5*spLength)
				DeltaDist = 0.5*spLength;

			return(Math.Sign(spDeltaLength) * Math.Pow(DeltaDist, 2) * spK);
		}

		private void CalculateTotalWallsForce(out double TotalXForce, out double TotalYForce)
		{
			double XResult = 0.0;
			double YResult = 0.0;

			double xForce, yForce;

			switch (ObsAvoidanceMethod)
			{
				case ObstacleAvoidanceMethods.Potential:
					foreach(Wall theWall in Environment.Walls)
					{
						ForceFromWall(theWall, out xForce, out yForce);

						XResult = XResult + xForce;
						YResult = YResult + yForce;
					}
					break;
				case ObstacleAvoidanceMethods.Sensor:
					foreach(Wall theWall in Environment.Walls)
					{
						double XVal, YVal;
						DistanceFromRectangle(
							this.AcX,
							this.AcY,
							theWall.X,
							theWall.Y,
							theWall.Width,
							theWall.Height,
							out XVal,
							out YVal);

						if (Math.Sqrt(Math.Pow(XVal,2) + Math.Pow(YVal, 2)) < MaxWallVisDist)
						{
							ForceFromWall(theWall, out xForce, out yForce);

							XResult = XResult + xForce;
							YResult = YResult + yForce;
						}
					}
					break;
			}

			TotalXForce = XResult;
			TotalYForce = YResult;
		}

		/// <summary>
		/// This routine of robot is called frequently (or the processor of mobile robot runs it itself)
		/// </summary>
		/// <param name="timerInterval">Time in milliseconds between two calls of this routine</param>
		public void Decide(int timerInterval)
		{
			AddingX = 0.0;
			AddingY = 0.0;

			double TotalXForce = 0.0;
			double TotalYForce = 0.0;

			double XVal = 0.0;
			double YVal = 0.0;

			Environment.RobotsInCircle(
				this.AcX,
				this.AcY,
				SeeDomain,
				ref VisibleRobots);

			switch (this.WorkMode)
			{
					#region WallFollowing
				case RobotWorkingModes.WallFollowing:					
					CalculateTotalWallsForce(out XVal, out YVal);
					
					DebugText = "X || Y == 0.0";
					if ((XVal < 0.0) && (YVal < 0.0))
					{
						DebugText = "(XVal < 0.0) && (YVal < 0.0)";
					}
					if ((XVal < 0.0) && (YVal > 0.0))
					{
						DebugText = "(XVal < 0.0) && (YVal > 0.0)";
					}
					if ((XVal > 0.0) && (YVal < 0.0))
					{
						DebugText = "(XVal > 0.0) && (YVal < 0.0)";
					}
					if ((XVal > 0.0) && (YVal > 0.0))
					{
						DebugText = "(XVal > 0.0) && (YVal > 0.0)";
					}

					if (XVal == 0.0)
					{
						if (YVal == 0.0)
						{
							DebugText = "(XVal == 0.0) && (YVal == 0)";
						}
						else
						{
							if (YVal < 0.0)
							{
								XVal = -1.0;
							}
							else
							{
								XVal = 1.0;
							}
						}
					}
					else
					{
						if (XVal < 0.0)
						{
							if (YVal == 0.0)
								YVal = -1.0;
						}
						else
						{
							if (XVal > 0.0)
							{
								if (YVal == 0.0)
									YVal = -1.0;
							}
						}
					}

					TotalXForce = TotalXForce + XVal;
					TotalYForce = TotalYForce + YVal;

					AddingX = Speed * TotalXForce * 0.0007;
					AddingY = Speed * TotalYForce * 0.0007;
					break;
					#endregion
					#region EnvironmentFollowing
				case RobotWorkingModes.EnvironmentFollowing:
					int CellX = (int)(this.AcX / Environment.CellsWidth);
					int CellY = (int)(this.AcY / Environment.CellsWidth);
					break;
					#endregion
					#region Leader
				case RobotWorkingModes.Leader:
					break;
					#endregion
					#region CompleteCoverage
				case RobotWorkingModes.CompleteCoverage:
//					ArrayList RobsInCircle = new ArrayList();
//					Environment.RobotsInCircle(
//						this.AcX,
//						this.AcY,
//						SeeDomain,
//						ref RobsInCircle);

					CalculateTotalWallsForce(out XVal, out YVal);
					TotalXForce = TotalXForce + XVal;
					TotalYForce = TotalYForce + YVal;

					foreach (MobileRobot rob in VisibleRobots/* RobsInCircle*/)
					{
						if (rob != this)
						{
							double Distance;
							Distance = Math.Sqrt(Math.Pow(rob.AcX-this.AcX,2) + Math.Pow(rob.AcY-this.AcY,2));

							double Force = 0.0;

							double SpringDeltaLength = SpringLength - Distance;
							if (Math.Abs(SpringDeltaLength) >= 0.1)
							{
//								Force = Math.Sign(SpringDeltaLength) * Math.Pow(SpringDeltaLength, 2) * this.Load * rob.Load;
								Force = SpringForce(SpringLength, SpringDeltaLength, this.Load * rob.Load);

								double XForce = - ((rob.AcX-this.AcX)*Force/Distance);
								double YForce = - ((rob.AcY-this.AcY)*Force/Distance);

								TotalXForce = TotalXForce + XForce;
								TotalYForce = TotalYForce + YForce;
							}
						}
					}

					AddingX = this.Speed * TotalXForce;
					AddingY = this.Speed * TotalYForce;
					break;
					#endregion
					#region TargetFollowing
				case RobotWorkingModes.TargetFollowing:
					CorrectPosition();

					if (ReachedGoal())
						return;

					foreach (Goal theGoal in Environment.Goals)
					{
						double Distance;
						Distance = Math.Sqrt(Math.Pow(theGoal.X-this.AcX,2) + Math.Pow(theGoal.Y-this.AcY,2));

						double Force;
						if (Distance != 0)
							Force = (1/Distance) * theGoal.Load * this.Load;
						else
							Force = 10;

						TotalXForce = TotalXForce - (theGoal.X-this.AcX)*(Force/Distance);
						TotalYForce = TotalYForce - (theGoal.Y-this.AcY)*(Force/Distance);
					}

					foreach (Obstacle theObstacle in Environment.Obstacles)
					{
						double Distance;
						Distance = Math.Sqrt(Math.Pow(theObstacle.X-this.AcX,2) + Math.Pow(theObstacle.Y-this.AcY,2));
						Distance = Distance - theObstacle.Radius - this.Radius;

						double Force;
						if (Distance != 0)
							Force = (1/Distance) * theObstacle.Load * this.Load;
						else
							Force = 1000;

						TotalXForce = TotalXForce - (theObstacle.X-this.AcX)*(Force/Distance);
						TotalYForce = TotalYForce - (theObstacle.Y-this.AcY)*(Force/Distance);
					}

					CalculateTotalWallsForce(out XVal, out YVal);
					TotalXForce = TotalXForce + XVal;
					TotalYForce = TotalYForce + YVal;

					foreach (MobileRobot theRob in Environment.MobileRobots)
					{
						if (theRob != this)
						{
							double Distance;
							Distance = Math.Sqrt(Math.Pow(theRob.AcX-this.AcX,2) + Math.Pow(theRob.AcY-this.AcY,2));
							//Distance = Distance - theRob.Radius - this.Radius;

							double Force;
							if (Distance != 0)
								Force = (1/Distance) * theRob.Load * this.Load;
							else
								Force = 10;

							TotalXForce = TotalXForce - (theRob.AcX-this.AcX)*(Force/Distance);
							TotalYForce = TotalYForce - (theRob.AcY-this.AcY)*(Force/Distance);
						}
					}
			
					AddingX = Speed * TotalXForce * 0.0007;
					AddingY = Speed * TotalYForce * 0.0007;
					break;
					#endregion
			}
		}

		private void CorrectPosition()
		{
//			ArrayList visRobots = new ArrayList();
//			Environment.RobotsInCircle(
//				AcX,
//				AcY,
//				SeeRadius,
//				ref visRobots);

			if (/*visRobots*/VisibleRobots.Count <= 1)
				return;

			double XError = 0;
			double YError = 0;
			double AverageXError = 0;
			double AverageYError = 0;
			double RelyOnOthersMult = 0.5;

			switch (CorrectionMethod)
			{
					#region None
				case RobotCorrectionMethods.None:
					break;
					#endregion
					#region Average
				case RobotCorrectionMethods.Average:
					foreach(MobileRobot visrob in VisibleRobots /*visRobots*/)
					{
						if (visrob != this)
						{
							double AcXDist = visrob.AcX - this.AcX;
							double AcYDist = visrob.AcY - this.AcY;
							double ImXDist = visrob.ImX - this.ImX;
							double ImYDist = visrob.ImY - this.ImY;

							XError = XError + AcXDist - ImXDist;
							YError = YError + AcYDist - ImYDist;
						}
					}

					AverageXError = XError/(VisibleRobots/*visRobots*/.Count-1);
					AverageYError = YError/(VisibleRobots/*visRobots*/.Count-1);

					RelyOnOthersMult = 1.0;

					this.ImX = (1 - RelyOnOthersMult) * this.ImX + RelyOnOthersMult * (this.ImX - AverageXError);
					this.ImY = (1 - RelyOnOthersMult) * this.ImY + RelyOnOthersMult * (this.ImY - AverageYError);
					break;
					#endregion
					#region WeightedAverage
				case RobotCorrectionMethods.WightedAverage:
					foreach(MobileRobot visrob in VisibleRobots /*visRobots*/)
					{
						if (visrob != this)
						{
							double AcDistance = Environment.Distance(
								this.AcX,
								this.AcY,
								visrob.AcX,
								visrob.AcY);
							double ImDistance = Environment.Distance(
								this.ImX,
								this.ImY,
								visrob.ImX,
								visrob.ImY);

							double ImXDist = visrob.ImX - this.ImX;
							double ImYDist = visrob.ImY - this.ImY;

							double EstAcXDist = (AcDistance * ImXDist) / ImDistance;
							double EstAcYDist = (AcDistance * ImYDist) / ImDistance;

							XError = XError + EstAcXDist - ImXDist;
							YError = YError + EstAcYDist - ImYDist;
						}
					}

					AverageXError = XError/(VisibleRobots/*visRobots*/.Count-1);
					AverageYError = YError/(VisibleRobots/*visRobots*/.Count-1);

					RelyOnOthersMult = 0.75;

					this.ImX = (1 - RelyOnOthersMult) * this.ImX + RelyOnOthersMult * (this.ImX - AverageXError);
					this.ImY = (1 - RelyOnOthersMult) * this.ImY + RelyOnOthersMult * (this.ImY - AverageYError);
					break;
					#endregion
					#region OWA
				case RobotCorrectionMethods.OWA:
					foreach(MobileRobot visrob in VisibleRobots /*visRobots*/)
					{
						if (visrob != this)
						{
							double AcXDist = visrob.AcX - this.AcX;
							double AcYDist = visrob.AcY - this.AcY;
							double ImXDist = visrob.ImX - this.ImX;
							double ImYDist = visrob.ImY - this.ImY;

							double AcDist = Math.Sqrt(Math.Pow(AcXDist,2) + Math.Pow(AcYDist,2));

							if (AcDist < 0.1)
							{
								XError = XError + AcXDist - ImXDist;
								YError = YError + AcYDist - ImYDist;
							}
							else
							{
								XError = XError + (0.003/AcDist) * (AcXDist - ImXDist);
								YError = YError + (0.003/AcDist) * (AcYDist - ImYDist);
							}
						}
					}

					AverageXError = XError/(/*visRobots*/VisibleRobots.Count-1);
					AverageYError = YError/(/*visRobots*/VisibleRobots.Count-1);

					RelyOnOthersMult = 0.5;

					this.ImX = (1 - RelyOnOthersMult) * this.ImX + RelyOnOthersMult * (this.ImX - AverageXError);
					this.ImY = (1 - RelyOnOthersMult) * this.ImY + RelyOnOthersMult * (this.ImY - AverageYError);
					break;
					#endregion
			}
		}

		/// <summary>
		/// Routine for moving the robot to a new point
		/// </summary>
		/// <param name="newX">Absolute value for X</param>
		/// <param name="newY">Absolute value for Y</param>
		public void MoveTo(double newXVal, double newYVal)
		{
			int LastCellX = (int)(this.AcX / Environment.CellsWidth);
			int LastCellY = (int)(this.AcY / Environment.CellsWidth);
			int NewCellX = (int)(newXVal / Environment.CellsWidth);
			int NewCellY = (int)(newYVal / Environment.CellsWidth);

			this.ImX = newXVal;
			this.ImY = newYVal;
			this.AcX = newXVal;
			this.AcY = newYVal;

			if ((NewCellX != LastCellX) || (NewCellY != LastCellY))
			{
				NumOfCellChanges++;
				Environment.DeclareRobotsPosition(this);
			}
		}

		private bool PosForRobotIsOK(double newX, double newY)
		{
			if (!((newX >= this.Radius) && (newX <= Environment.GroundWidth)))
				return (false);

			if (!((newY >= this.Radius) && (newY <= Environment.GroundHeight)))
				return(false);

			foreach(Wall theWall in Environment.Walls)
				if (PointInRectangle(
					newX,
					newY,
					theWall.X,
					theWall.Y,
					theWall.Width,
					theWall.Height))
					return(false);

			return(true);
		}

		public void Move(double xChange, double yChange)
		{
			double NewImXVal = this.ImX + xChange;
			double NewImYVal = this.ImY + yChange;
				
			if (PosForRobotIsOK(NewImXVal, NewImYVal))
			{
				this.ImX = NewImXVal;
				this.ImY = NewImYVal;
			}

			double N1, N2;
			Environment.NormalRandomNumber(out N1, out N2);
			
			double XRandVal = ErrorMultiplier * xChange * (N1);
			double YRandVal = ErrorMultiplier * yChange * (N2);

			double NewAcXVal = this.AcX + xChange + XRandVal;
			double NewAcYVal = this.AcY + yChange + YRandVal;

			if (PosForRobotIsOK(NewAcXVal, NewAcYVal))
			{
				int LastCellX = (int)(this.AcX / Environment.CellsWidth);
				int LastCellY = (int)(this.AcY / Environment.CellsWidth);
				int NewCellX = (int)(NewAcXVal / Environment.CellsWidth);
				int NewCellY = (int)(NewAcYVal / Environment.CellsWidth);

				this.AcX = NewAcXVal;
				this.AcY = NewAcYVal;

				if ((NewCellX != LastCellX) || (NewCellY != LastCellY))
				{
					NumOfCellChanges++;
					Environment.DeclareRobotsPosition(this);
				}
			}
		}

		public void Do(int timerInterval)
		{
			Move(AddingX, AddingY);
		}

		public double GetRealX()
		{
			return AcX;
		}

		public double GetRealY()
		{
			return AcY;
		}
	}
}
